import pathlib as pl
import os

style_dir = pl.Path(__file__).resolve().parent

futura = style_dir / 'futura_presentation.mplstyle'
breve = style_dir / 'breve_presentation.mplstyle'
default = style_dir / 'presentation.mplstyle'